import abc
import typing as ta

from .. import check
from .. import lang
from .base import AnyLifecycle  # noqa
from .base import Lifecycle  # noqa
from .states import LifecycleState
from .states import LifecycleStates
from .transitions import LifecycleTransition
from .transitions import LifecycleTransitions


R = ta.TypeVar('R')

AnyLifecycleT = ta.TypeVar('AnyLifecycleT', bound='AnyLifecycle')

LifecycleT = ta.TypeVar('LifecycleT', bound='Lifecycle')


##


class AnyLifecycleListener(ta.Generic[AnyLifecycleT, R]):
    def on_starting(self, obj: AnyLifecycleT) -> R | None:
        return None

    def on_started(self, obj: AnyLifecycleT) -> R | None:
        return None

    def on_stopping(self, obj: AnyLifecycleT) -> R | None:
        return None

    def on_stopped(self, obj: AnyLifecycleT) -> R | None:
        return None


class AnyLifecycleController(AnyLifecycle[R], lang.Abstract, ta.Generic[AnyLifecycleT, R]):
    def __init__(
            self,
            lifecycle: AnyLifecycleT,
    ) -> None:
        super().__init__()

        self._lifecycle: AnyLifecycleT = check.isinstance(lifecycle, AnyLifecycle)  # type: ignore

        self._state = LifecycleStates.NEW
        self._listeners: list[AnyLifecycleListener[AnyLifecycleT, R]] = []

    __repr__ = lang.attr_ops(lambda o: (o.lifecycle, o.state)).repr

    @property
    def lifecycle(self) -> AnyLifecycleT:
        return self._lifecycle

    @property
    def state(self) -> LifecycleState:
        return self._state

    def add_listener(self, listener: AnyLifecycleListener[AnyLifecycleT, R]) -> ta.Self:
        self._listeners.append(check.isinstance(listener, AnyLifecycleListener))
        return self

    @abc.abstractmethod
    def _advance(
            self,
            transition: LifecycleTransition,
            lifecycle_fn: ta.Callable[[], R | None],
            pre_listener_fn: ta.Callable[
                [AnyLifecycleListener[AnyLifecycleT, R]],
                ta.Callable[[AnyLifecycleT], R | None],
            ] | None = None,
            post_listener_fn: ta.Callable[
                [AnyLifecycleListener[AnyLifecycleT, R]],
                ta.Callable[[AnyLifecycleT], R | None],
            ] | None = None,
    ) -> R | None:
        raise NotImplementedError

    ##

    @ta.override
    def lifecycle_construct(self) -> R | None:
        return self._advance(
            LifecycleTransitions.CONSTRUCT,
            self._lifecycle.lifecycle_construct,
        )

    @ta.override
    def lifecycle_start(self) -> R | None:
        return self._advance(
            LifecycleTransitions.START,
            self._lifecycle.lifecycle_start,
            lambda l: l.on_starting,
            lambda l: l.on_started,
        )

    @ta.override
    def lifecycle_stop(self) -> R | None:
        return self._advance(
            LifecycleTransitions.STOP,
            self._lifecycle.lifecycle_stop,
            lambda l: l.on_stopping,
            lambda l: l.on_stopped,
        )

    @ta.override
    def lifecycle_destroy(self) -> R | None:
        return self._advance(
            LifecycleTransitions.DESTROY,
            self._lifecycle.lifecycle_destroy,
        )


##


LifecycleListener: ta.TypeAlias = AnyLifecycleListener[LifecycleT, None]


class LifecycleController(
    AnyLifecycleController[LifecycleT, None],
    Lifecycle,
    ta.Generic[LifecycleT],
):
    def __init__(
            self,
            lifecycle: LifecycleT,
            *,
            lock: lang.DefaultLockable = None,
    ) -> None:
        super().__init__(lifecycle)

        self._lock = lang.default_lock(lock, False)

    def _advance(
            self,
            transition: LifecycleTransition,
            lifecycle_fn: ta.Callable[[], None],
            pre_listener_fn: ta.Callable[
                [LifecycleListener[LifecycleT]],
                ta.Callable[[LifecycleT], None],
            ] | None = None,
            post_listener_fn: ta.Callable[
                [LifecycleListener[LifecycleT]],
                ta.Callable[[LifecycleT], None],
            ] | None = None,
    ) -> None:
        with self._lock():
            if pre_listener_fn is not None:
                for listener in self._listeners:
                    pre_listener_fn(listener)(self._lifecycle)
            check.state(self._state in transition.old)
            self._state = transition.new_intermediate
            try:
                lifecycle_fn()
            except Exception:
                self._state = transition.new_failed
                raise
            self._state = transition.new_succeeded
            if post_listener_fn is not None:
                for listener in self._listeners:
                    post_listener_fn(listener)(self._lifecycle)



from .errors import (  # noqa
    KnownError,
    KnownErrors,

    CUSTOM_ERROR_BASE,
)

from .types import (  # noqa
    NUMBER_TYPES,
    Number,
    Object,
    ID_TYPES,
    Id,

    VERSION,

    NotSpecified,
    is_not_specified,
    check_not_not_specified,

    Request,
    request,
    notification,

    Response,
    result,

    Error,
    error,

    Message,
    detect_message_type,
)


##


from ... import marshal as _msh

_msh.register_global_module_import('._marshal', __package__)



import typing as ta

from ... import lang
from .. import json
from ..codecs import make_object_lazy_loaded_codec
from ..codecs import make_str_object_codec


if ta.TYPE_CHECKING:
    from . import parsing
    from . import rendering
else:
    parsing = lang.proxy_import('.parsing', __package__)
    rendering = lang.proxy_import('.rendering', __package__)


##


def dumps(
        obj: ta.Any,
        *,
        multiline_strings: bool = False,
        **kwargs: ta.Any,
) -> str:
    if multiline_strings:
        return rendering.Json5Renderer.render_str(
            obj,
            multiline_strings=True,
            **kwargs,
        )

    else:
        return json.dumps(obj, **kwargs)


def dumps_pretty(obj: ta.Any, **kwargs: ta.Any) -> str:
    return dumps(obj, **json.PRETTY_KWARGS, **kwargs)


def dumps_compact(obj: ta.Any, **kwargs: ta.Any) -> str:
    return dumps(obj, **json.COMPACT_KWARGS, **kwargs)


def dump(obj: ta.Any, fp: ta.Any, **kwargs: ta.Any) -> None:
    fp.write(dumps(obj, **kwargs))


def dump_pretty(obj: ta.Any, fp: ta.Any, **kwargs: ta.Any) -> None:
    fp.write(dumps_pretty(obj, **kwargs))


def dump_compact(obj: ta.Any, fp: ta.Any, **kwargs: ta.Any) -> None:
    fp.write(dumps_compact(obj, **kwargs))


#


def loads(s: str, **kwargs: ta.Any) -> ta.Any:
    return parsing.parse(s, **kwargs)


def load(fp: ta.Any, **kwargs: ta.Any) -> ta.Any:
    return loads(fp.read(), **kwargs)


##


JSON5_CODEC = make_str_object_codec('json5', dumps, loads)

# @omlish-manifest
_JSON5_LAZY_CODEC = make_object_lazy_loaded_codec(__name__, 'JSON5_CODEC', JSON5_CODEC)



import functools
import typing as ta


T = ta.TypeVar('T')

CallableT = ta.TypeVar('CallableT', bound=ta.Callable)


##


class _AbstractCachedNullary:
    def __init__(self, fn):
        super().__init__()

        self._fn = fn
        self._value = self._missing = object()
        functools.update_wrapper(self, fn)

    def __call__(self, *args, **kwargs):  # noqa
        raise TypeError

    def __get__(self, instance, owner=None):  # noqa
        bound = instance.__dict__[self._fn.__name__] = self.__class__(self._fn.__get__(instance, owner))
        return bound


##


class _CachedNullary(_AbstractCachedNullary):
    def __call__(self, *args, **kwargs):  # noqa
        if self._value is self._missing:
            self._value = self._fn()
        return self._value


def cached_nullary(fn: CallableT) -> CallableT:
    return _CachedNullary(fn)  # type: ignore


def static_init(fn: CallableT) -> CallableT:
    fn = cached_nullary(fn)
    fn()
    return fn


##


class _AsyncCachedNullary(_AbstractCachedNullary):
    async def __call__(self, *args, **kwargs):
        if self._value is self._missing:
            self._value = await self._fn()
        return self._value


def async_cached_nullary(fn):  # ta.Callable[..., T]) -> ta.Callable[..., T]:
    return _AsyncCachedNullary(fn)


##


cached_property = functools.cached_property


class _cached_property:  # noqa
    """Backported to pick up https://github.com/python/cpython/commit/056dfc71dce15f81887f0bd6da09d6099d71f979 ."""

    def __init__(self, func):
        self.func = func
        self.attrname = None  # noqa
        self.__doc__ = func.__doc__
        self.__module__ = func.__module__

    _NOT_FOUND = object()

    def __set_name__(self, owner, name):
        if self.attrname is None:
            self.attrname = name  # noqa
        elif name != self.attrname:
            raise TypeError(
                f'Cannot assign the same cached_property to two different names ({self.attrname!r} and {name!r}).',
            )

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        if self.attrname is None:
            raise TypeError('Cannot use cached_property instance without calling __set_name__ on it.')

        try:
            cache = instance.__dict__
        except AttributeError:  # not all objects have __dict__ (e.g. class defines slots)
            raise TypeError(
                f"No '__dict__' attribute on {type(instance).__name__!r} instance to cache {self.attrname!r} property.",
            ) from None

        val = cache.get(self.attrname, self._NOT_FOUND)

        if val is self._NOT_FOUND:
            val = self.func(instance)
            try:
                cache[self.attrname] = val
            except TypeError:
                raise TypeError(
                    f"The '__dict__' attribute on {type(instance).__name__!r} instance does not support item "
                    f"assignment for caching {self.attrname!r} property.",
                ) from None

        return val


globals()['cached_property'] = _cached_property



"""
TODO:
 - explicit blacklisted key element - blacklist lite Injector
"""
import dataclasses as dc
import typing as ta

from .. import check
from .. import reflect as rfl
from ..lite import inject as lij
from .bindings import Binding
from .elements import Elements
from .keys import Key
from .providers import ConstProvider
from .providers import CtorProvider
from .providers import FnProvider
from .providers import LinkProvider
from .providers import Provider
from .scopes import Singleton
from .types import Scope
from .types import Unscoped


##


@dc.dataclass(frozen=True)
class BoxedLiteProvider:
    p: Provider
    sc: Scope | None = None


##


@ta.overload
def convert_from_lite(o: lij.InjectorBindings) -> Elements:
    ...


@ta.overload
def convert_from_lite(o: lij.InjectorKey) -> Key:
    ...


@ta.overload
def convert_from_lite(o: lij.InjectorBinding) -> Binding:
    ...


@ta.overload
def convert_from_lite(o: lij.InjectorProvider) -> BoxedLiteProvider:
    ...


def convert_from_lite(o):
    if isinstance(o, lij.InjectorBindings):
        return Elements([
            convert_from_lite(c)
            for c in o.bindings()
        ])

    elif isinstance(o, lij.InjectorKey):
        check.not_equal(o.cls_, lij.InjectorKeyCls)
        check.arg(not o.array)
        return Key(rfl.type_(o.cls_), tag=o.tag)

    elif isinstance(o, lij.InjectorBinding):
        blp = convert_from_lite(o.provider)
        return Binding(
            convert_from_lite(o.key),
            blp.p,
            blp.sc if blp.sc is not None else Unscoped(),
        )

    elif isinstance(o, lij.InjectorProvider):
        if isinstance(o, lij.FnInjectorProvider):
            return BoxedLiteProvider(FnProvider(
                o.fn,
            ))

        elif isinstance(o, lij.CtorInjectorProvider):
            return BoxedLiteProvider(CtorProvider(
                o.cls_,
            ))

        elif isinstance(o, lij.ConstInjectorProvider):
            return BoxedLiteProvider(ConstProvider(
                o.v,
            ))

        elif isinstance(o, lij.SingletonInjectorProvider):
            blp = convert_from_lite(o.p)
            check.none(blp.sc)
            return dc.replace(
                blp,
                sc=Singleton(),
            )

        elif isinstance(o, lij.LinkInjectorProvider):
            return BoxedLiteProvider(LinkProvider(
                convert_from_lite(o.k),
            ))

        elif isinstance(o, lij.ScopedInjectorProvider):
            raise NotImplementedError

        else:
            raise TypeError(o)

    else:
        raise TypeError(o)



# ruff: noqa: UP007
# @omlish-lite
import logging
import os
import sys
import threading
import time
import typing as ta

from ...logs.infos import LoggingContextInfos
from ...logs.levels import LogLevel
from .types import ABSENT_TYPED_LOGGER_VALUE
from .types import AbsentTypedLoggerValue
from .types import DefaultTypedLoggerValue
from .types import TypedLoggerValue


##


class StandardTypedLoggerValues:
    def __new__(cls, *args, **kwargs):  # noqa
        raise TypeError

    #

    class TimeNs(DefaultTypedLoggerValue[int]):
        @classmethod
        def _default_value(cls) -> float:
            return time.time()

    class Time(DefaultTypedLoggerValue[float]):
        _default_key = True
        _default_value = TypedLoggerValue.ContextLambda(
            lambda ctx: ctx[StandardTypedLoggerValues.TimeNs].map_absent(
                lambda tv: tv.v / 1e9,
            ),
        )

    #

    class Level(TypedLoggerValue[LogLevel]):
        pass

    class LevelName(DefaultTypedLoggerValue[str]):
        _default_key = 'level'
        _default_value = TypedLoggerValue.ContextLambda(
            lambda ctx: ctx[StandardTypedLoggerValues.Level].map_absent(
                lambda lvl: logging.getLevelName(lvl.v),
            ),
        )

    #

    class Msg(TypedLoggerValue[str]):
        _default_key = True

    #

    class Caller(TypedLoggerValue[LoggingContextInfos.Caller]):
        pass

    #

    class Tid(DefaultTypedLoggerValue[int]):
        @classmethod
        def _default_value(cls) -> ta.Union[int, AbsentTypedLoggerValue]:
            if hasattr(threading, 'get_native_id'):
                return threading.get_native_id()
            else:
                return ABSENT_TYPED_LOGGER_VALUE

    class ThreadIdent(DefaultTypedLoggerValue[int]):
        @classmethod
        def _default_value(cls) -> int:
            return threading.get_ident()

    class ThreadName(DefaultTypedLoggerValue[int]):
        @classmethod
        def _default_value(cls) -> str:
            return threading.current_thread().name

    #

    class Pid(DefaultTypedLoggerValue[int]):
        @classmethod
        def _default_value(cls) -> int:
            return os.getpid()

    class ProcessName(DefaultTypedLoggerValue[str]):
        @classmethod
        def _default_value(cls) -> ta.Union[str, AbsentTypedLoggerValue]:
            if (mp := sys.modules.get('multiprocessing')) is None:
                return ABSENT_TYPED_LOGGER_VALUE

            # Errors may occur if multiprocessing has not finished loading yet - e.g. if a custom import hook causes
            # third-party code to run when multiprocessing calls import. See issue 8200 for an example
            try:
                return mp.current_process().name
            except Exception:  # noqa
                return ABSENT_TYPED_LOGGER_VALUE

    #

    class AsyncioTaskName(DefaultTypedLoggerValue[str]):
        @classmethod
        def _default_value(cls) -> ta.Union[str, AbsentTypedLoggerValue]:
            if (asyncio := sys.modules.get('asyncio')) is None:
                return ABSENT_TYPED_LOGGER_VALUE

            try:
                return asyncio.current_task().get_name()
            except Exception:  # noqa
                return ABSENT_TYPED_LOGGER_VALUE



import typing as ta

from .base import Kv
from .base import MutableKv


K = ta.TypeVar('K')
V = ta.TypeVar('V')


##


class MappingKv(Kv[K, V]):
    def __init__(self, m: ta.Mapping[K, V]) -> None:
        super().__init__()

        self._m = m

    def __getitem__(self, k: K, /) -> V:
        return self._m[k]

    def __len__(self) -> int:
        return len(self._m)

    def items(self) -> ta.Iterator[tuple[K, V]]:
        return iter(self._m.items())


class MappingMutableKv(MappingKv[K, V], MutableKv[K, V]):
    def __init__(self, m: ta.MutableMapping[K, V]) -> None:
        super().__init__(m)

    _m: ta.MutableMapping[K, V]

    def __setitem__(self, k: K, v: V, /) -> None:
        self._m[k] = v

    def __delitem__(self, k: K, /) -> None:
        del self._m[k]


##


class KvMapping(ta.Mapping[K, V]):
    def __init__(self, kv: Kv[K, V]) -> None:
        super().__init__()

        self._kv = kv

    def __getitem__(self, key: K, /) -> V:
        return self._kv[key]

    def __len__(self) -> int:
        return len(self._kv)

    # FIXME: ItemsView
    def items(self) -> ta.Iterator[tuple[K, V]]:  # type: ignore[override]
        return iter(self._kv.items())

    def __iter__(self) -> ta.Iterator[K]:
        return (k for k, v in self.items())


class KvMutableMapping(KvMapping[K, V], ta.MutableMapping[K, V]):
    def __init__(self, kv: MutableKv[K, V]) -> None:
        super().__init__(kv)

    _kv: MutableKv[K, V]

    def __setitem__(self, key: K, value: V, /) -> None:
        self._kv[key] = value

    def __delitem__(self, key: K, /) -> None:
        del self._kv[key]



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
"""
TODO:
 - use ..text.mangle
"""


##


def mangle_path(path: str) -> str:
    if not path.startswith('/'):
        raise ValueError('Only absolute Unix paths are supported')
    return (
        path
        .replace('_', '__')
        .replace('/', '_')
    )


def unmangle_path(mangled: str) -> str:
    if '/' in mangled:
        raise ValueError("Mangled paths should not contain '/'")
    return (
        mangled
        .replace('_', '/')
        .replace('//', '_')
    )



"""
TODO:
 - -x / --exec - os.exec entrypoint
 - refuse to install non-exec-relevant Bootstraps when chosen
"""
# ruff: noqa: UP006 UP007 UP045
import argparse
import dataclasses as dc
import io
import itertools
import os
import shutil
import sys
import typing as ta

from .. import lang
from .base import Bootstrap
from .harness import BOOTSTRAP_TYPES_BY_NAME
from .harness import bootstrap


if ta.TYPE_CHECKING:
    import runpy

else:
    runpy = lang.proxy_import('runpy')


##


class _OrderedArgsAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        if 'ordered_args' not in namespace:
            setattr(namespace, 'ordered_args', [])
        if self.const is not None:
            value = self.const
        else:
            value = values
        namespace.ordered_args.append((self.dest, value))


def _or_opt(ty):
    return (ty, ta.Optional[ty])


def _int_or_str(v):
    try:
        return int(v)
    except ValueError:
        return v


def _add_arguments(parser: argparse.ArgumentParser) -> None:
    # ta.Optional[ta.Mapping[str, ta.Tuple[ta.Optional[int], ta.Optional[int]]]]

    for cname, cls in BOOTSTRAP_TYPES_BY_NAME.items():
        for fld in dc.fields(cls.Config):
            aname = f'--{cname}:{fld.name}'
            kw: ta.Dict[str, ta.Any] = {}

            if fld.type in _or_opt(str):
                pass
            elif fld.type in _or_opt(bool):
                kw.update(const=True, nargs=0)
            elif fld.type in _or_opt(int):
                kw.update(type=int)
            elif fld.type in _or_opt(float):
                kw.update(type=float)
            elif fld.type in _or_opt(ta.Union[int, str]):
                kw.update(type=_int_or_str)

            elif fld.type in (
                    *_or_opt(ta.Sequence[str]),
                    *_or_opt(ta.Mapping[str, ta.Optional[str]]),
                    *_or_opt(ta.Mapping[int, ta.Union[int, str, None]]),
                    *_or_opt(ta.Mapping[str, ta.Tuple[ta.Optional[int], ta.Optional[int]]]),
            ):
                if aname[-1] != 's':
                    raise NameError(aname)
                aname = aname[:-1]

            else:
                raise TypeError(fld)

            parser.add_argument(aname.replace('_', '-'), action=_OrderedArgsAction, **kw)


def _process_arguments(args: ta.Any) -> ta.Sequence[Bootstrap.Config]:
    if not (oa := getattr(args, 'ordered_args', None)):
        return []

    cfgs: ta.List[Bootstrap.Config] = []

    for cname, cargs in [
        (n, list(g))
        for n, g in
        itertools.groupby(oa, key=lambda s: s[0].partition(':')[0])
    ]:
        ccls = BOOTSTRAP_TYPES_BY_NAME[cname].Config
        flds = {f.name: f for f in dc.fields(ccls)}

        kw: ta.Dict[str, ta.Any] = {}
        for aname, aval in cargs:
            k = aname.partition(':')[2]

            if k not in flds:
                k += 's'
                fld = flds[k]

                if fld.type in _or_opt(ta.Sequence[str]):
                    kw.setdefault(k, []).append(aval)

                elif fld.type in _or_opt(ta.Mapping[str, ta.Optional[str]]):
                    if '=' not in aval:
                        kw.setdefault(k, {})[aval] = None
                    else:
                        ek, _, ev = aval.partition('=')
                        kw.setdefault(k, {})[ek] = ev

                elif fld.type in _or_opt(ta.Mapping[int, ta.Union[int, str, None]]):
                    fk, _, fv = aval.partition('=')
                    if not fv:
                        kw.setdefault(k, {})[int(fk)] = None
                    else:
                        kw.setdefault(k, {})[int(fk)] = _int_or_str(fv)

                elif fld.type in _or_opt(ta.Mapping[str, ta.Tuple[ta.Optional[int], ta.Optional[int]]]):
                    fk, _, fv = aval.partition('=')
                    if ',' in fv:
                        tl, tr = fv.split(',')
                    else:
                        tl, tr = None, None
                    kw.setdefault(k, {})[fk] = (_int_or_str(tl) if tl else None, _int_or_str(tr) if tr else None)

                else:
                    raise TypeError(fld)

            else:
                kw[k] = aval

        cfg = ccls(**kw)
        cfgs.append(cfg)

    return cfgs


##


def _main() -> int:
    parser = argparse.ArgumentParser()

    _add_arguments(parser)

    parser.add_argument('-m', '--module', action='store_true')
    parser.add_argument('-c', '--code', action='store_true')
    parser.add_argument('-x', '--exec', action='store_true')
    parser.add_argument('target')
    parser.add_argument('args', nargs=argparse.REMAINDER)

    args = parser.parse_args()

    if len([a for a in ('module', 'code', 'exec') if getattr(args, a)]) > 1:
        raise Exception('Multiple exec mode specified')

    cfgs = _process_arguments(args)

    with bootstrap(*cfgs):
        tgt = args.target

        if args.exec:
            exe = shutil.which(tgt)
            if exe is None:
                raise FileNotFoundError(exe)

            os.execl(exe, exe, *args.args)

            return 0  # type: ignore  # noqa

        sys.argv = [tgt, *(args.args or ())]

        if args.code:
            exec(tgt, {}, {})

        elif args.module:
            runpy._run_module_as_main(tgt)  # type: ignore  # noqa

        else:
            with io.open_code(tgt) as fp:
                src = fp.read()

            ns = dict(
                __name__='__main__',
                __file__=tgt,
                __builtins__=__builtins__,
                __spec__=None,
            )

            import __main__  # noqa
            __main__.__dict__.clear()
            __main__.__dict__.update(ns)
            exec(compile(src, tgt, 'exec'), __main__.__dict__, __main__.__dict__)

        return 0


if __name__ == '__main__':
    sys.exit(_main())



"""
TODO:
 - sha1 is slow :/ key by repr but name by sha1
"""
import dataclasses as dc
import re
import typing as ta

from .... import lang
from .base import Plan


##


@dc.dataclass(frozen=True)
class Plans:
    tup: tuple[Plan, ...]

    def __iter__(self) -> ta.Iterator[Plan]:
        return iter(self.tup)

    @lang.cached_function(no_wrapper_update=True)
    def render(self) -> str:
        return _render(self)


##


_WS_PAT = re.compile(r'\s+')


def _render(plans: Plans) -> str:
    return _WS_PAT.sub('', repr(plans.tup))



"""
TODO:
 - standardize following redirects
"""
import contextlib
import functools
import typing as ta

from ... import dataclasses as dc
from ... import lang
from ...io.buffers import ReadableListBuffer
from ..headers import HttpHeaders
from .asyncs import AsyncHttpClient
from .asyncs import AsyncStreamHttpResponse
from .base import HttpClientContext
from .base import HttpClientError
from .base import HttpRequest
from .sync import HttpClient
from .sync import StreamHttpResponse


if ta.TYPE_CHECKING:
    import httpx
else:
    httpx = lang.proxy_import('httpx')


##


class HttpxHttpClient(HttpClient):
    @dc.dataclass(frozen=True)
    class _StreamAdapter:
        it: ta.Iterator[bytes]
        buf: ReadableListBuffer = dc.field(default_factory=ReadableListBuffer)

        def read1(self, /, n: int = -1) -> bytes:
            if n < 0:
                if (b := self.buf.read(n)) is not None:
                    return b
                try:
                    return next(self.it)
                except StopIteration:
                    return b''

            else:
                while len(self.buf) < n:
                    try:
                        b = next(self.it)
                    except StopIteration:
                        b = b''
                    if not b:
                        return self.buf.read() or b''
                    self.buf.feed(b)
                return self.buf.read(n) or b''

    def _stream_request(self, ctx: HttpClientContext, req: HttpRequest) -> StreamHttpResponse:
        try:
            resp_cm = httpx.stream(
                method=req.method_or_default,
                url=req.url,
                headers=req.headers_ or None,  # type: ignore
                content=req.data,
                timeout=req.timeout_s,
            )

        except httpx.HTTPError as e:
            raise HttpClientError from e

        resp_close = functools.partial(resp_cm.__exit__, None, None, None)

        try:
            resp = resp_cm.__enter__()
            return StreamHttpResponse(
                status=resp.status_code,
                headers=HttpHeaders(resp.headers.raw),
                request=req,
                underlying=resp,
                stream=self._StreamAdapter(resp.iter_bytes()),
                _closer=resp_close,  # type: ignore
            )

        except httpx.HTTPError as e:
            resp_close()
            raise HttpClientError from e

        except Exception:
            resp_close()
            raise


##


class HttpxAsyncHttpClient(AsyncHttpClient):
    @dc.dataclass(frozen=True)
    class _StreamAdapter:
        it: ta.AsyncIterator[bytes]
        buf: ReadableListBuffer = dc.field(default_factory=ReadableListBuffer)

        async def read1(self, /, n: int = -1) -> bytes:
            if n < 0:
                if (b := self.buf.read(n)) is not None:
                    return b
                try:
                    return await anext(self.it)
                except StopAsyncIteration:
                    return b''

            else:
                while len(self.buf) < n:
                    try:
                        b = await anext(self.it)
                    except StopAsyncIteration:
                        b = b''
                    if not b:
                        return self.buf.read() or b''
                    self.buf.feed(b)
                return self.buf.read(n) or b''

    async def _stream_request(self, ctx: HttpClientContext, req: HttpRequest) -> AsyncStreamHttpResponse:
        es = contextlib.AsyncExitStack()

        try:
            client = await es.enter_async_context(httpx.AsyncClient())

            resp = await es.enter_async_context(client.stream(
                method=req.method_or_default,
                url=req.url,
                headers=req.headers_ or None,  # type: ignore
                content=req.data,
                timeout=req.timeout_s,
            ))

            return AsyncStreamHttpResponse(
                status=resp.status_code,
                headers=HttpHeaders(resp.headers.raw),
                request=req,
                underlying=resp,
                stream=self._StreamAdapter(resp.aiter_bytes()),
                _closer=es.aclose,
            )

        except httpx.HTTPError as e:
            await es.aclose()
            raise HttpClientError from e

        except Exception:
            await es.aclose()
            raise



# @omlish-lite
# @omlish-amalg ../../omdev/scripts/lib/logs.py
from .base import AnyLogger  # noqa
from .standard import configure_standard_logging  # noqa
from .std.loggers import StdLogger  # noqa


##



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
"""
TODO:
 - def parse: (<bind>)?:<port>, unix://, fd://
 - unix chown/chgrp
 - DupSocketBinder
 - udp
"""
import abc
import dataclasses as dc
import errno
import os
import socket as socket_
import stat
import typing as ta

from ..lite.abstract import Abstract
from ..lite.check import check
from ..lite.dataclasses import dataclass_maybe_post_init
from .addresses import SocketAddress
from .addresses import SocketAndAddress


SocketBinderT = ta.TypeVar('SocketBinderT', bound='SocketBinder')
SocketBinderConfigT = ta.TypeVar('SocketBinderConfigT', bound='SocketBinder.Config')

CanSocketBinderConfig = ta.Union['SocketBinder.Config', int, ta.Tuple[str, int], str]  # ta.TypeAlias
CanSocketBinder = ta.Union['SocketBinder', CanSocketBinderConfig]  # ta.TypeAlias


##


class SocketBinder(Abstract, ta.Generic[SocketBinderConfigT]):
    @dc.dataclass(frozen=True)
    class Config:
        listen_backlog: int = 5

        allow_reuse_address: bool = True
        allow_reuse_port: bool = True

        set_inheritable: bool = False

        #

        @classmethod
        def of(cls, obj: CanSocketBinderConfig) -> 'SocketBinder.Config':
            if isinstance(obj, SocketBinder.Config):
                return obj

            elif isinstance(obj, int):
                return TcpSocketBinder.Config(
                    port=obj,
                )

            elif isinstance(obj, tuple):
                host, port = obj
                return TcpSocketBinder.Config(
                    host=host,
                    port=port,
                )

            elif isinstance(obj, str):
                return UnixSocketBinder.Config(
                    file=obj,
                )

            else:
                raise TypeError(obj)

    #

    def __init__(self, config: SocketBinderConfigT) -> None:
        super().__init__()

        self._config = config

    #

    @classmethod
    def of(cls, obj: CanSocketBinder) -> 'SocketBinder':
        if isinstance(obj, SocketBinder):
            return obj

        config: SocketBinder.Config
        if isinstance(obj, SocketBinder.Config):
            config = obj

        else:
            config = SocketBinder.Config.of(obj)

        if isinstance(config, TcpSocketBinder.Config):
            return TcpSocketBinder(config)

        elif isinstance(config, UnixSocketBinder.Config):
            return UnixSocketBinder(config)

        else:
            raise TypeError(config)

    #

    class Error(RuntimeError):
        pass

    class NotBoundError(Error):
        pass

    class AlreadyBoundError(Error):
        pass

    #

    @property
    @abc.abstractmethod
    def address_family(self) -> int:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def address(self) -> SocketAddress:
        raise NotImplementedError

    #

    _socket: socket_.socket

    @property
    def is_bound(self) -> bool:
        return hasattr(self, '_socket')

    @property
    def socket(self) -> socket_.socket:
        try:
            return self._socket
        except AttributeError:
            raise self.NotBoundError from None

    _name: str

    @property
    def name(self) -> str:
        try:
            return self._name
        except AttributeError:
            raise self.NotBoundError from None

    _port: ta.Optional[int]

    @property
    def port(self) -> ta.Optional[int]:
        try:
            return self._port
        except AttributeError:
            raise self.NotBoundError from None

    #

    def fileno(self) -> int:
        return self.socket.fileno()

    #

    def __enter__(self: SocketBinderT) -> SocketBinderT:
        self.bind()

        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.close()

    #

    def _init_socket(self) -> None:
        if hasattr(self, '_socket'):
            raise self.AlreadyBoundError

        sock = socket_.socket(self.address_family, socket_.SOCK_STREAM)
        self._socket = sock

        if self._config.allow_reuse_address and hasattr(socket_, 'SO_REUSEADDR'):
            sock.setsockopt(socket_.SOL_SOCKET, socket_.SO_REUSEADDR, 1)

        # Since Linux 6.12.9, SO_REUSEPORT is not allowed on other address families than AF_INET/AF_INET6.
        if (
                self._config.allow_reuse_port and hasattr(socket_, 'SO_REUSEPORT') and
                self.address_family in (socket_.AF_INET, socket_.AF_INET6)
        ):
            try:
                sock.setsockopt(socket_.SOL_SOCKET, socket_.SO_REUSEPORT, 1)
            except OSError as err:
                if err.errno not in (errno.ENOPROTOOPT, errno.EINVAL):
                    raise

        if self._config.set_inheritable and hasattr(sock, 'set_inheritable'):
            sock.set_inheritable(True)

    def _pre_bind(self) -> None:
        pass

    def _post_bind(self) -> None:
        pass

    def bind(self) -> None:
        self._init_socket()

        self._pre_bind()

        self.socket.bind(self.address)

        self._post_bind()

        check.state(all(hasattr(self, a) for a in ('_socket', '_name', '_port')))

    #

    def close(self) -> None:
        if hasattr(self, '_socket'):
            self._socket.close()

    #

    def listen(self) -> None:
        self.socket.listen(self._config.listen_backlog)

    @abc.abstractmethod
    def accept(self, sock: ta.Optional[socket_.socket] = None) -> SocketAndAddress:
        raise NotImplementedError


##


class TcpSocketBinder(SocketBinder):
    @dc.dataclass(frozen=True)
    class Config(SocketBinder.Config):
        DEFAULT_HOST: ta.ClassVar[str] = 'localhost'
        host: str = DEFAULT_HOST

        port: int = 0

        def __post_init__(self) -> None:
            dataclass_maybe_post_init(super())
            check.non_empty_str(self.host)
            check.isinstance(self.port, int)
            check.arg(self.port > 0)

    def __init__(self, config: Config) -> None:
        super().__init__(check.isinstance(config, self.Config))

        self._address = (config.host, config.port)

    #

    address_family = socket_.AF_INET

    @property
    def address(self) -> SocketAddress:
        return self._address

    #

    def _post_bind(self) -> None:
        super()._post_bind()

        host, port, *_ = self.socket.getsockname()

        self._name = socket_.getfqdn(host)
        self._port = port

    #

    def accept(self, sock: ta.Optional[socket_.socket] = None) -> SocketAndAddress:
        if sock is None:
            sock = self.socket

        conn, client_address = sock.accept()
        return SocketAndAddress(conn, client_address)


##


class UnixSocketBinder(SocketBinder):
    @dc.dataclass(frozen=True)
    class Config(SocketBinder.Config):
        file: str = ''

        unlink: bool = False

        def __post_init__(self) -> None:
            dataclass_maybe_post_init(super())
            check.non_empty_str(self.file)

    def __init__(self, config: Config) -> None:
        super().__init__(check.isinstance(config, self.Config))

        self._address = config.file

    #

    address_family = socket_.AF_UNIX

    @property
    def address(self) -> SocketAddress:
        return self._address

    #

    def _pre_bind(self) -> None:
        super()._pre_bind()

        if self._config.unlink:
            try:
                if stat.S_ISSOCK(os.stat(self._config.file).st_mode):
                    os.unlink(self._config.file)
            except FileNotFoundError:
                pass

    def _post_bind(self) -> None:
        super()._post_bind()

        name = self.socket.getsockname()

        os.chmod(name, stat.S_IRWXU | stat.S_IRWXG)  # noqa

        self._name = name
        self._port = None

    #

    def accept(self, sock: ta.Optional[socket_.socket] = None) -> SocketAndAddress:
        if sock is None:
            sock = self.socket

        conn, _ = sock.accept()
        client_address = ('', 0)
        return SocketAndAddress(conn, client_address)



"""
TODO:
 - sync vs async
"""
import abc
import typing as ta

from ... import lang
from .columns import Column
from .columns import Columns
from .queries import Query
from .resources import ContextCloser
from .rows import Row


##


class Querier(ContextCloser, lang.Abstract):
    @property
    @abc.abstractmethod
    def adapter(self) -> 'Adapter':
        raise NotImplementedError

    @abc.abstractmethod
    def query(self, query: Query) -> 'Rows':  # ta.Raises[QueryError]
        raise NotImplementedError


##


class Rows(ContextCloser, ta.Iterator[Row], lang.Abstract):
    @property
    @abc.abstractmethod
    def columns(self) -> Columns:
        raise NotImplementedError

    @ta.final
    def __iter__(self) -> ta.Self:
        return self

    @abc.abstractmethod
    def __next__(self) -> Row:  # ta.Raises[StopIteration]
        raise NotImplementedError


##


class Conn(Querier, lang.Abstract):
    pass


##


class Db(Querier, lang.Abstract):
    @abc.abstractmethod
    def connect(self) -> Conn:
        raise NotImplementedError

    def query(self, query: Query) -> Rows:
        with self.connect() as conn:
            return conn.query(query)


##


class Adapter(lang.Abstract):
    @abc.abstractmethod
    def scan_type(self, c: Column) -> type:
        raise NotImplementedError



import abc
import typing as ta

from ... import lang
from ... import reflect as rfl
from ...funcs import guard as gfs
from .configs import ConfigRegistry
from .contexts import MarshalContext
from .contexts import MarshalFactoryContext
from .contexts import UnmarshalContext
from .contexts import UnmarshalFactoryContext
from .values import Value


T = ta.TypeVar('T')


##


class Marshaler(lang.Abstract):
    @abc.abstractmethod
    def marshal(self, ctx: MarshalContext, o: ta.Any) -> Value:
        raise NotImplementedError


class Unmarshaler(lang.Abstract):
    @abc.abstractmethod
    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Any:
        raise NotImplementedError


##


MarshalerMaker: ta.TypeAlias = gfs.GuardFn[[MarshalFactoryContext, rfl.Type], Marshaler]
UnmarshalerMaker: ta.TypeAlias = gfs.GuardFn[[UnmarshalFactoryContext, rfl.Type], Unmarshaler]


class MarshalerFactory(lang.Abstract):
    @abc.abstractmethod
    def make_marshaler(self, ctx: MarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], Marshaler] | None:
        raise NotImplementedError


class UnmarshalerFactory(lang.Abstract):
    @abc.abstractmethod
    def make_unmarshaler(self, ctx: UnmarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], Unmarshaler] | None:
        raise NotImplementedError


##


class Marshaling(lang.Abstract):
    @abc.abstractmethod
    def config_registry(self) -> ConfigRegistry:
        raise NotImplementedError

    @abc.abstractmethod
    def marshaler_factory(self) -> MarshalerFactory:
        raise NotImplementedError

    @abc.abstractmethod
    def unmarshaler_factory(self) -> UnmarshalerFactory:
        raise NotImplementedError

    ##

    def new_marshal_factory_context(self) -> MarshalFactoryContext:
        return MarshalFactoryContext(
            configs=self.config_registry(),
            marshaler_factory=self.marshaler_factory(),
        )

    def new_unmarshal_factory_context(self) -> UnmarshalFactoryContext:
        return UnmarshalFactoryContext(
            configs=self.config_registry(),
            unmarshaler_factory=self.unmarshaler_factory(),
        )

    ##

    def new_marshal_context(self) -> MarshalContext:
        return MarshalContext(
            configs=self.config_registry(),
            marshal_factory_context=self.new_marshal_factory_context(),
        )

    def new_unmarshal_context(self) -> UnmarshalContext:
        return UnmarshalContext(
            configs=self.config_registry(),
            unmarshal_factory_context=self.new_unmarshal_factory_context(),
        )

    #

    @ta.final
    def marshal(self, obj: ta.Any, ty: ta.Any | None = None, **kwargs: ta.Any) -> Value:
        return self.new_marshal_context(**kwargs).marshal(obj, ty)

    @ta.overload
    def unmarshal(self, v: Value, ty: type[T], **kwargs: ta.Any) -> T:
        ...

    @ta.overload
    def unmarshal(self, v: Value, ty: ta.Any, **kwargs: ta.Any) -> ta.Any:
        ...

    @ta.final
    def unmarshal(self, v, ty, **kwargs):
        return self.new_unmarshal_context(**kwargs).unmarshal(v, ty)


##


# MarshalerOrUnmarshaler: ta.TypeAlias = Marshaler | Unmarshaler
# MarshalerOrUnmarshalerT = ta.TypeVar('MarshalerOrUnmarshalerT', bound=MarshalerOrUnmarshaler)
#
# MarshalContextOrUnmarshalContext: ta.TypeAlias = MarshalContext | UnmarshalContext
# MarshalContextOrUnmarshalContextT = ta.TypeVar('MarshalContextOrUnmarshalContextT', bound=MarshalContextOrUnmarshalContext)  # noqa
#
# MarshalerMakerOrUnmarshalerMaker: ta.TypeAlias = MarshalerMaker | UnmarshalerMaker



# ruff: noqa: UP006
import abc
import typing as ta


##


_ABSTRACT_METHODS_ATTR = '__abstractmethods__'
_IS_ABSTRACT_METHOD_ATTR = '__isabstractmethod__'


def is_abstract_method(obj: ta.Any) -> bool:
    return bool(getattr(obj, _IS_ABSTRACT_METHOD_ATTR, False))


def update_abstracts(cls, *, force=False):
    if not force and not hasattr(cls, _ABSTRACT_METHODS_ATTR):
        # Per stdlib: We check for __abstractmethods__ here because cls might by a C implementation or a python
        # implementation (especially during testing), and we want to handle both cases.
        return cls

    abstracts: ta.Set[str] = set()

    for scls in cls.__bases__:
        for name in getattr(scls, _ABSTRACT_METHODS_ATTR, ()):
            value = getattr(cls, name, None)
            if getattr(value, _IS_ABSTRACT_METHOD_ATTR, False):
                abstracts.add(name)

    for name, value in cls.__dict__.items():
        if getattr(value, _IS_ABSTRACT_METHOD_ATTR, False):
            abstracts.add(name)

    setattr(cls, _ABSTRACT_METHODS_ATTR, frozenset(abstracts))
    return cls


#


class AbstractTypeError(TypeError):
    pass


_FORCE_ABSTRACT_ATTR = '__forceabstract__'


class Abstract:
    """
    Different from, but interoperable with, abc.ABC / abc.ABCMeta:

     - This raises AbstractTypeError during class creation, not instance instantiation - unless Abstract or abc.ABC are
       explicitly present in the class's direct bases.
     - This will forbid instantiation of classes with Abstract in their direct bases even if there are no
       abstractmethods left on the class.
     - This is a mixin, not a metaclass.
     - As it is not an ABCMeta, this does not support virtual base classes. As a result, operations like `isinstance`
       and `issubclass` are ~7x faster.
     - It additionally enforces a base class order of (Abstract, abc.ABC) to preemptively prevent common mro conflicts.

    If not mixed-in with an ABCMeta, it will update __abstractmethods__ itself.
    """

    __slots__ = ()

    __abstractmethods__: ta.ClassVar[ta.FrozenSet[str]] = frozenset()

    #

    def __forceabstract__(self):
        raise TypeError

    # This is done manually, rather than through @abc.abstractmethod, to mask it from static analysis.
    setattr(__forceabstract__, _IS_ABSTRACT_METHOD_ATTR, True)

    #

    def __init_subclass__(cls, **kwargs: ta.Any) -> None:
        setattr(
            cls,
            _FORCE_ABSTRACT_ATTR,
            getattr(Abstract, _FORCE_ABSTRACT_ATTR) if Abstract in cls.__bases__ else False,
        )

        super().__init_subclass__(**kwargs)

        if not (Abstract in cls.__bases__ or abc.ABC in cls.__bases__):
            ams = {a: cls for a, o in cls.__dict__.items() if is_abstract_method(o)}

            seen = set(cls.__dict__)
            for b in cls.__bases__:
                ams.update({a: b for a in set(getattr(b, _ABSTRACT_METHODS_ATTR, [])) - seen})  # noqa
                seen.update(dir(b))

            if ams:
                raise AbstractTypeError(
                    f'Cannot subclass abstract class {cls.__name__} with abstract methods: ' +
                    ', '.join(sorted([
                        '.'.join([
                            *([m] if (m := getattr(c, '__module__')) else []),
                            getattr(c, '__qualname__', getattr(c, '__name__')),
                            a,
                        ])
                        for a, c in ams.items()
                    ])),
                )

        xbi = (Abstract, abc.ABC)  # , ta.Generic ?
        bis = [(cls.__bases__.index(b), b) for b in xbi if b in cls.__bases__]
        if bis != sorted(bis):
            raise TypeError(
                f'Abstract subclass {cls.__name__} must have proper base class order of '
                f'({", ".join(getattr(b, "__name__") for b in xbi)}), got: '
                f'({", ".join(getattr(b, "__name__") for _, b in sorted(bis))})',
            )

        if not isinstance(cls, abc.ABCMeta):
            update_abstracts(cls, force=True)



from .. import lang as _lang


with _lang.auto_proxy_init(globals()):
    ##

    from .executors import (  # noqa
        ImmediateExecutor,
        new_executor,
    )

    from .futures import (  # noqa
        FutureError,
        FutureTimeoutError,
        wait_futures,
        wait_dependent_futures,
    )



"""
TODO:
 - vs ta.get_type_hints?
  - eval (ta.get_type_hints) vs forwardref (inspect.get_annotations)?
 - remove, or at least address, get_filtered_type_hints wrt 3.14
"""
import typing as ta

from .. import check
from .. import lang
from ..lite.typing import typing_annotations_attr


if ta.TYPE_CHECKING:
    import inspect
else:
    inspect = lang.proxy_import('inspect')


annotationlib = lang.lazy_import('annotationlib', optional=True, cache_failure=True)


##


def has_annotations(obj: ta.Any) -> bool:
    return hasattr(obj, typing_annotations_attr())


##


def get_annotations(obj: ta.Any) -> ta.Mapping[str, ta.Any]:
    if (al := annotationlib()) is not None:
        return al.get_annotations(obj, format=al.Format.FORWARDREF)  # noqa
    else:
        return inspect.get_annotations(obj)


##


class _TypeHintsTarget:
    def __init__(self, obj, annotations):
        super().__init__()

        self.obj = obj
        self.__annotations__ = annotations
        self.__globals__ = obj.__globals__
        self.__type_params__ = obj.__type_params__


def get_filtered_type_hints(
        obj: ta.Any,
        *,
        include: ta.Container[str] | None = None,
        exclude: ta.Container[str] | None = None,
) -> ta.Mapping[str, ta.Any]:
    """
    Gross hack: get_type_hints doesn't support recursive types, but it's nice to support functions with, at least,
    recursive return types.

    This is consciously constrained in what it supports - basically just functions.
    """

    check.not_isinstance(include, str)
    check.not_isinstance(exclude, str)

    anns = {
        k: v
        for k, v in obj.__annotations__.items()
        if (include is None or k in include)
        and not (exclude is not None and k in exclude)
    }

    return ta.get_type_hints(_TypeHintsTarget(obj, anns))



import typing as ta

from ... import dataclasses as dc
from ..types import Tag


##


@dc.dataclass(frozen=True)
@dc.extra_class_params(cache_hash=True, terse_repr=True)
class Id:
    """A utility dataclass intended to be used as a key tag for disambiguation."""

    v: int

    tag: ta.Any = dc.xfield(
        default=None,
        kw_only=True,
        validate=lambda o: not isinstance(o, Tag),
        repr_fn=dc.opt_repr,
    )



import typing as ta


T = ta.TypeVar('T')


##


class OrderedSet(ta.MutableSet[T]):
    def __init__(self, iterable: ta.Iterable[T] | None = None) -> None:
        super().__init__()

        self._dct: dict[T, ta.Any] = {}
        if iterable is not None:
            self |= iterable  # type: ignore  # noqa

    def __len__(self) -> int:
        return len(self._dct)

    def __contains__(self, item: ta.Any) -> bool:
        return item in self._dct

    def add(self, item: T) -> None:
        if item not in self._dct:
            self._dct[item] = None

    def update(self, items: ta.Iterable[T]) -> None:
        for item in items:
            if item not in self._dct:
                self._dct[item] = None

    def discard(self, item: T) -> None:
        if item in self._dct:
            del self._dct[item]

    def __iter__(self) -> ta.Iterator[T]:
        return iter(self._dct.keys())

    def __reversed__(self):
        return reversed(self._dct.keys())

    def pop(self, last: bool = True) -> T:
        if not self:
            raise KeyError('set is empty')
        item = next(reversed(self._dct.keys()))
        self.discard(item)
        return item

    def __repr__(self) -> str:
        if not self:
            return f'{self.__class__.__name__}()'
        return f'{self.__class__.__name__}({list(self)!r})'

    def __eq__(self, other) -> bool:
        if isinstance(other, OrderedSet):
            return len(self) == len(other) and list(self) == list(other)
        return set(self) == set(other)

    def __ne__(self, other) -> bool:
        return not (self == other)

    def __hash__(self) -> int:
        raise TypeError


class OrderedFrozenSet(ta.FrozenSet[T]):  # noqa
    _list: ta.Sequence[T]

    def __new__(cls, items: ta.Iterable[T]) -> frozenset[T]:  # type: ignore
        item_set = set()
        item_list = []
        for item in items:
            if item not in item_set:
                item_set.add(item)
                item_list.append(item)
        obj = super(cls, OrderedFrozenSet).__new__(cls, item_set)
        obj._list = item_list  # type: ignore  # noqa
        return obj

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}([{", ".join(map(repr, self))}])'

    def __iter__(self) -> ta.Iterator[T]:
        return iter(self._list)

    def __sub__(self, other: ta.Iterable[T]) -> frozenset[T]:
        s = set(other)
        return type(self)(i for i in self if i not in s)



import abc

from ... import lang
from ..coro import BytesSteppedCoro
from ..coro import BytesSteppedReaderCoro


##


class Compression(lang.Abstract):
    @abc.abstractmethod
    def compress(self, d: bytes) -> bytes:
        raise NotImplementedError

    @abc.abstractmethod
    def decompress(self, d: bytes) -> bytes:
        raise NotImplementedError


class IncrementalCompression(lang.Abstract):
    @abc.abstractmethod
    def compress_incremental(self) -> BytesSteppedCoro[None]:
        raise NotImplementedError

    @abc.abstractmethod
    def decompress_incremental(self) -> BytesSteppedReaderCoro[None]:
        raise NotImplementedError



"""
TODO:
 - col.TypeMap?
  - at least get_any
"""
import abc
import dataclasses as dc
import threading
import typing as ta

from ... import lang


##


class RegistryItem(lang.Abstract):
    pass


RegistryItemT = ta.TypeVar('RegistryItemT', bound=RegistryItem)
RegistryItemU = ta.TypeVar('RegistryItemU', bound=RegistryItem)


##


class RegistryView(lang.Abstract, ta.Generic[RegistryItemT]):
    @abc.abstractmethod
    def get(
            self,
            key: ta.Any,
            *,
            identity: bool | None = None,
    ) -> ta.Sequence[RegistryItemT]:
        ...

    @abc.abstractmethod
    def get_of(
            self,
            key: ta.Any,
            item_ty: type[RegistryItemU],
            *,
            identity: bool | None = None,
    ) -> ta.Sequence[RegistryItemU]:
        ...


##


class RegistrySealedError(Exception):
    pass


class Registry(RegistryView[RegistryItemT]):
    def __init__(
            self,
            *,
            lock: ta.Optional[threading.RLock] = None,  # noqa
    ) -> None:
        super().__init__()

        if lock is None:
            lock = threading.RLock()
        self._lock = lock

        self._state: Registry._State[RegistryItemT] = Registry._State(
            dct={},
            id_dct={},
            version=0,
        )

        self._sealed = False

    #

    @dc.dataclass(frozen=True)
    class _KeyItems(ta.Generic[RegistryItemU]):
        key: ta.Any
        items: ta.Sequence[RegistryItemU] = ()
        item_lists_by_ty: ta.Mapping[type[RegistryItemU], ta.Sequence[RegistryItemU]] = dc.field(default_factory=dict)

        def add(self, *items: RegistryItemU) -> 'Registry._KeyItems[RegistryItemU]':
            item_lists_by_ty: dict[type[RegistryItemU], list[RegistryItemU]] = {}

            for i in items:
                try:
                    l = item_lists_by_ty[type(i)]
                except KeyError:
                    l = item_lists_by_ty[type(i)] = list(self.item_lists_by_ty.get(type(i), ()))
                l.append(i)

            return Registry._KeyItems(
                self.key,
                (*self.items, *items),
                {**self.item_lists_by_ty, **item_lists_by_ty},
            )

    @dc.dataclass(frozen=True, kw_only=True)
    class _State(ta.Generic[RegistryItemU]):
        dct: ta.Mapping[ta.Any, 'Registry._KeyItems[RegistryItemU]']
        id_dct: ta.Mapping[ta.Any, 'Registry._KeyItems[RegistryItemU]']
        version: int

        #

        def register(
                self,
                key: ta.Any,
                *items: RegistryItemT,
                identity: bool = False,
        ) -> 'Registry._State[RegistryItemU]':
            if not items:
                return self

            sr_dct: ta.Any = self.dct if not identity else self.id_dct
            if (sr := sr_dct.get(key)) is None:
                sr = Registry._KeyItems(key)
            sr = sr.add(*items)
            new = {key: sr}

            return Registry._State(
                dct={**self.dct, **new} if not identity else self.dct,
                id_dct={**self.id_dct, **new} if identity else self.id_dct,
                version=self.version + 1,
            )

        #

        _get_cache: dict[ta.Any, ta.Sequence[RegistryItem]] = dc.field(default_factory=dict)

        def get(
                self,
                key: ta.Any,
                *,
                identity: bool | None = None,
        ) -> ta.Sequence[RegistryItem]:
            if identity is None:
                try:
                    return self._get_cache[key]
                except KeyError:
                    pass

                ret = self._get_cache[key] = (
                    *self.get(key, identity=True),
                    *self.get(key, identity=False),
                )
                return ret

            dct: ta.Any = self.dct if not identity else self.id_dct
            try:
                return dct[key].items
            except KeyError:
                return ()

        _get_of_cache: dict[ta.Any, dict[type, ta.Sequence[RegistryItem]]] = dc.field(default_factory=dict)

        def get_of(
                self,
                key: ta.Any,
                item_ty: type[RegistryItem],
                *,
                identity: bool | None = None,
        ) -> ta.Sequence[RegistryItem]:
            if identity is None:
                try:
                    return self._get_of_cache[key][item_ty]
                except KeyError:
                    pass

                ret = self._get_of_cache.setdefault(key, {})[item_ty] = (
                    *self.get_of(key, item_ty, identity=True),
                    *self.get_of(key, item_ty, identity=False),
                )
                return ret

            dct: ta.Any = self.dct if not identity else self.id_dct
            try:
                sr = dct[key]
            except KeyError:
                return ()
            return sr.item_lists_by_ty.get(item_ty, ())

    def is_sealed(self) -> bool:
        if self._sealed:
            return True
        with self._lock:
            return self._sealed

    def seal(self) -> ta.Self:
        if self._sealed:
            raise RegistrySealedError(self)
        with self._lock:
            self._seal()
        return self

    def _seal(self) -> None:
        if self._sealed:
            raise RegistrySealedError(self)

        self._sealed = True

    #

    def register(
            self,
            key: ta.Any,
            *items: RegistryItemT,
            identity: bool = False,
    ) -> ta.Self:
        if not items:
            return self

        with self._lock:
            if self._sealed:
                raise RegistrySealedError(self)

            self._state = self._state.register(
                key,
                *items,
                identity=identity,
            )

        return self

    #

    def get(
            self,
            key: ta.Any,
            *,
            identity: bool | None = None,
    ) -> ta.Sequence[RegistryItemT]:
        return self._state.get(key, identity=identity)  # type: ignore [return-value]

    def get_of(
            self,
            key: ta.Any,
            item_ty: type[RegistryItemU],
            *,
            identity: bool | None = None,
    ) -> ta.Sequence[RegistryItemU]:
        return self._state.get_of(key, item_ty, identity=identity)  # type: ignore [return-value]



import abc
import enum
import functools
import os
import sys
import threading
import typing as ta

from .. import check
from .. import dataclasses as dc
from .. import lang
from ..diag import pydevd
from .targets import Target


if ta.TYPE_CHECKING:
    import multiprocessing as mp
    import multiprocessing.context
    import multiprocessing.process  # noqa

    from ..multiprocessing import spawn as omp_spawn

else:
    mp = lang.proxy_import('multiprocessing', extras=['context', 'process'])
    subprocess = lang.proxy_import('subprocess')

    omp_spawn = lang.proxy_import('..multiprocessing.spawn', __package__)


##


class Spawning(dc.Case):
    pass


class Spawn(dc.Frozen, final=True):
    fn: ta.Callable[[], None]

    _: dc.KW_ONLY

    target: Target | None = None

    inherit_fds: ta.Collection[int] | None = None


class Spawner(lang.ContextManaged, lang.Abstract):
    @abc.abstractmethod
    def spawn(self, spawn: Spawn) -> None:
        raise NotImplementedError


class InProcessSpawner(Spawner, lang.Abstract):
    pass


@functools.singledispatch
def spawner_for(spawning: Spawning) -> Spawner:
    raise TypeError(spawning)


##


class MultiprocessingSpawning(Spawning, kw_only=True):
    class StartMethod(enum.Enum):
        SPAWN = enum.auto()
        FORK = enum.auto()
        # TODO: FORK_SERVER

    # Defaults to 'fork' if under pydevd, else 'spawn'
    start_method: StartMethod | None = None

    #

    # Note: Per multiprocessing docs, `no_linger=True` processes (corresponding to `Process(daemon=True)`) cannot spawn
    # subprocesses, and thus will fail if `Daemon.Config.reparent_process` is set.
    no_linger: bool = False

    #

    @dc.dataclass(frozen=True, kw_only=True)
    class EntrypointArgs:
        spawning: 'MultiprocessingSpawning'
        spawn: Spawn
        start_method: 'MultiprocessingSpawning.StartMethod'

    entrypoint: ta.Callable[[EntrypointArgs], None] | None = None


class MultiprocessingSpawner(Spawner):
    def __init__(self, spawning: MultiprocessingSpawning) -> None:
        super().__init__()

        self._spawning = spawning
        self._process: ta.Optional['mp.process.BaseProcess'] = None  # noqa

    @lang.cached_function
    def _determine_start_method(self) -> 'MultiprocessingSpawning.StartMethod':
        if (start_method := self._spawning.start_method) is not None:
            return start_method

        # Unfortunately, pydevd forces the use of the 'fork' start_method, which cannot be mixed with 'spawn':
        #   https://github.com/python/cpython/blob/a7427f2db937adb4c787754deb4c337f1894fe86/Lib/multiprocessing/spawn.py#L102  # noqa
        if pydevd.is_running():
            return MultiprocessingSpawning.StartMethod.FORK

        return MultiprocessingSpawning.StartMethod.SPAWN

    def _process_cls(self, spawn: Spawn) -> type['mp.process.BaseProcess']:
        start_method = self._determine_start_method()

        ctx: 'mp.context.BaseContext'  # noqa
        if start_method == MultiprocessingSpawning.StartMethod.FORK:
            ctx = mp.get_context(check.non_empty_str('fork'))

        elif start_method == MultiprocessingSpawning.StartMethod.SPAWN:
            ctx = omp_spawn.ExtrasSpawnContext(omp_spawn.SpawnExtras(
                pass_fds=frozenset(spawn.inherit_fds) if spawn.inherit_fds is not None else None,
            ))

        else:
            raise ValueError(start_method)

        return ctx.Process  # type: ignore

    def spawn(self, spawn: Spawn) -> None:
        check.none(self._process)

        target: ta.Callable[[], None]
        if (ep := self._spawning.entrypoint) is not None:
            target = functools.partial(ep, MultiprocessingSpawning.EntrypointArgs(
                spawning=self._spawning,
                spawn=spawn,
                start_method=self._determine_start_method(),
            ))
        else:
            target = spawn.fn

        self._process = self._process_cls(spawn)(
            target=target,
            daemon=self._spawning.no_linger,
        )
        self._process.start()


@spawner_for.register
def _(spawning: MultiprocessingSpawning) -> MultiprocessingSpawner:
    return MultiprocessingSpawner(spawning)


##


class ForkSpawning(Spawning):
    pass


class ForkSpawner(Spawner, dc.Frozen):
    spawning: ForkSpawning

    def spawn(self, spawn: Spawn) -> None:
        if (pid := os.fork()):  # noqa
            return

        try:
            spawn.fn()
        except BaseException:  # noqa
            sys.exit(1)
        else:
            sys.exit(0)

        raise RuntimeError('Unreachable')  # noqa


@spawner_for.register
def _(spawning: ForkSpawning) -> ForkSpawner:
    return ForkSpawner(spawning)


##


class ThreadSpawning(Spawning, kw_only=True):
    linger: bool = False


class ThreadSpawner(InProcessSpawner):
    def __init__(self, spawning: ThreadSpawning) -> None:
        super().__init__()

        self._spawning = spawning
        self._thread: threading.Thread | None = None

    def spawn(self, spawn: Spawn) -> None:
        check.none(self._thread)
        self._thread = threading.Thread(
            target=spawn.fn,
            daemon=not self._spawning.linger,
        )
        self._thread.start()


@spawner_for.register
def _(spawning: ThreadSpawning) -> ThreadSpawner:
    return ThreadSpawner(spawning)



# @omlish-lite
import typing as ta

from .levels import LogLevel


##


@ta.runtime_checkable
class LoggerLike(ta.Protocol):
    """Satisfied by both our Logger and stdlib logging.Logger."""

    def isEnabledFor(self, level: LogLevel) -> bool: ...  # noqa

    def getEffectiveLevel(self) -> LogLevel: ...  # noqa

    #

    def log(self, level: LogLevel, msg: str, /, *args: ta.Any, **kwargs: ta.Any) -> None: ...  # noqa

    def debug(self, msg: str, /, *args: ta.Any, **kwargs: ta.Any) -> None: ...  # noqa

    def info(self, msg: str, /, *args: ta.Any, **kwargs: ta.Any) -> None: ...  # noqa

    def warning(self, msg: str, /, *args: ta.Any, **kwargs: ta.Any) -> None: ...  # noqa

    def error(self, msg: str, /, *args: ta.Any, **kwargs: ta.Any) -> None: ...  # noqa

    def exception(self, msg: str, /, *args: ta.Any, **kwargs: ta.Any) -> None: ...  # noqa

    def critical(self, msg: str, /, *args: ta.Any, **kwargs: ta.Any) -> None: ...  # noqa



from ._registry import register


##


PARAM_NAME = '__repeat'


@register
class RepeatPlugin:
    def pytest_addoption(self, parser):
        parser.addoption('--repeat', action='store', type=int, help='Number of times to repeat each test')

    def pytest_generate_tests(self, metafunc):
        if metafunc.config.option.repeat is None:
            return

        n = metafunc.config.option.repeat
        metafunc.fixturenames.append(PARAM_NAME)
        metafunc.parametrize(PARAM_NAME, range(n))



from .... import lang
from .base import KnownKeyword
from .base import StrKeyword


##


class FormatKeyword(KnownKeyword, lang.Abstract, lang.Sealed):
    pass


##


class Format(StrKeyword, FormatKeyword, lang.Final, tag='format'):
    pass



import dataclasses as dc
import sys
import typing as ta


T = ta.TypeVar('T')

A0 = ta.TypeVar('A0')
A1 = ta.TypeVar('A1')
A2 = ta.TypeVar('A2')


##
# A workaround for typing deficiencies (like `Argument 2 to NewType(...) must be subclassable`).
#
# Note that this problem doesn't happen at runtime - it happens in mypy:
#
#   mypy <(echo "import typing as ta; MyCallback = ta.NewType('MyCallback', ta.Callable[[], None])")
#   /dev/fd/11:1:22: error: Argument 2 to NewType(...) must be subclassable (got "Callable[[], None]")  [valid-newtype]
#


@dc.dataclass(frozen=True)
class AnyFunc(ta.Generic[T]):
    fn: ta.Callable[..., T]

    def __call__(self, *args: ta.Any, **kwargs: ta.Any) -> T:
        return self.fn(*args, **kwargs)


@dc.dataclass(frozen=True)
class Func0(ta.Generic[T]):
    fn: ta.Callable[[], T]

    def __call__(self) -> T:
        return self.fn()


@dc.dataclass(frozen=True)
class Func1(ta.Generic[A0, T]):
    fn: ta.Callable[[A0], T]

    def __call__(self, a0: A0) -> T:
        return self.fn(a0)


@dc.dataclass(frozen=True)
class Func2(ta.Generic[A0, A1, T]):
    fn: ta.Callable[[A0, A1], T]

    def __call__(self, a0: A0, a1: A1) -> T:
        return self.fn(a0, a1)


@dc.dataclass(frozen=True)
class Func3(ta.Generic[A0, A1, A2, T]):
    fn: ta.Callable[[A0, A1, A2], T]

    def __call__(self, a0: A0, a1: A1, a2: A2) -> T:
        return self.fn(a0, a1, a2)


##


_TYPING_ANNOTATIONS_ATTR = '__annotate__' if sys.version_info >= (3, 14) else '__annotations__'


def typing_annotations_attr() -> str:
    return _TYPING_ANNOTATIONS_ATTR



import dataclasses as dc
import typing as ta
import weakref

from .. import check
from .. import lang
from .mappings import dict_factory


T = ta.TypeVar('T')
K = ta.TypeVar('K')
V = ta.TypeVar('V')


##


class AttrRegistry(ta.Generic[K, V]):
    """
    MRO-honoring class member registry. There are many ways to do this, and this one is attr name based: a class is
    considered to have objects registered to it based on whether or not they are accessible by a non-shadowed,
    MRO-resolved named attribute on that class.

    Care must be taken when overriding registered objects from superclasses in subclasses - shadowing the attribute name
    of the superclass member will not automatically register the new member with the same name to the registry - it must
    be explicitly registered itself. This is a feature, allowing for selective de-registration of objects in subclasses
    via name shadowing.
    """

    def __init__(
            self,
            *,
            requires_override: bool = False,
            is_override: ta.Callable[[ta.Any], bool] | None = None,
            forbid_duplicates: bool = False,
            identity: bool = False,
            weak: bool = False,
    ) -> None:
        super().__init__()

        self._requires_override = requires_override
        self._is_override = is_override
        self._forbid_duplicates = forbid_duplicates
        self._identity = identity
        self._weak = weak

        self._objs: ta.MutableMapping[K, V] = dict_factory(identity=identity, weak=weak)()
        self._invalidate_callbacks: list[ta.Callable[[], None]] = []

    def add_invalidate_callback(self, callback: ta.Callable[[], None]) -> None:
        self._invalidate_callbacks.append(callback)

    def register(self, obj: K, val: V) -> None:
        check.not_in(obj, self._objs)

        self._objs[obj] = val

        for iv in self._invalidate_callbacks:
            iv()

    def _lookup(self, obj: ta.Any) -> lang.Maybe[V]:
        if not self._identity:
            try:
                hash(obj)
            except TypeError:
                return lang.empty()

        try:
            val = self._objs[obj]
        except KeyError:
            return lang.empty()
        else:
            return lang.just(val)

    @dc.dataclass()
    class DuplicatesForbiddenError(Exception):
        owner_cls: type
        instance_cls: type
        att: str
        ex_att: str

    def collect(self, instance_cls: type, owner_cls: type | None = None) -> dict[str, tuple[K, V]]:
        if owner_cls is None:
            owner_cls = instance_cls

        mro = instance_cls.__mro__[-2::-1]
        try:
            mro_pos = mro.index(owner_cls)
        except ValueError:
            raise TypeError(f'Owner class {owner_cls} not in mro of instance class {instance_cls}') from None

        mro_dct: dict[str, list[tuple[type, ta.Any]]] = {}
        for cur_cls in mro[:mro_pos + 1]:
            for att, obj in cur_cls.__dict__.items():
                if att not in mro_dct:
                    if not self._lookup(obj).present:
                        continue

                try:
                    lst = mro_dct[att]
                except KeyError:
                    lst = mro_dct[att] = []
                lst.append((cur_cls, obj))

        #

        seen: ta.MutableMapping[ta.Any, str] | None = None
        if self._forbid_duplicates:
            seen = dict_factory(identity=self._identity)()

        out: dict[str, tuple[K, V]] = {}
        for att, lst in mro_dct.items():
            if not lst:
                raise RuntimeError
            _, obj = lst[-1]

            if len(lst) > 1:
                if self._requires_override and not (self._is_override or lang.is_override)(obj):
                    raise lang.RequiresOverrideError(
                        att,
                        instance_cls,
                        lst[-1][0],
                        lst[0][0],
                    )

            if not (mv := self._lookup(obj)).present:
                continue

            if seen is not None:
                try:
                    ex_att = seen[obj]
                except KeyError:
                    pass
                else:
                    raise AttrRegistry.DuplicatesForbiddenError(owner_cls, instance_cls, att, ex_att)  # noqa
                seen[obj] = att

            out[att] = (obj, mv.must())

        return out


##


class AttrRegistryCache(ta.Generic[K, V, T]):
    def __init__(
            self,
            registry: AttrRegistry[K, V],
            prepare: ta.Callable[[type, dict[str, tuple[K, V]]], T],
    ) -> None:
        super().__init__()

        self._registry = registry
        self._prepare = prepare

        self._cache: dict[ta.Any, T] = {}

        def cache_remove(k, self_ref=weakref.ref(self)):
            if (ref_self := self_ref()) is not None:
                cache = ref_self._cache  # noqa
                try:
                    del cache[k]
                except KeyError:
                    pass

        self._cache_remove = cache_remove

        registry.add_invalidate_callback(self._cache.clear)

    def get(self, instance_cls: type) -> T:
        cls_ref = weakref.ref(instance_cls)
        try:
            return self._cache[cls_ref]
        except KeyError:
            pass
        del cls_ref

        collected = self._registry.collect(instance_cls)
        out = self._prepare(instance_cls, collected)
        self._cache[weakref.ref(instance_cls, self._cache_remove)] = out
        return out
